from torch.nn import Module, ModuleList, ModuleDict, Sequential

from torch.nn import Parameter

# from torch import init
# from torch import utils

from torch.nn import MaxPool1d, MaxPool2d, MaxPool3d
from torch.nn import AvgPool1d, AvgPool2d, AvgPool3d

from .activation import ReLU, ReLU6
from .batchnorm import BatchNorm2d
from .conv import Conv2d
from .linear import Linear

from . import extra
from .adder import Adder2d

from .number import qsigned, qunsigned


from .sd import threshold_weight_3sd, threshold_bias_3sd, threshold_activation_3sd
from .kl import entropy_calibration
from .max import threshold_activation_max, threshold_bias_max, threshold_weight_max
from .hook import hook_handler, remove_hook, remove_hookout, add_hook, get_hook
from .init import init_network, qsigned


from .max import *
from .sd import *
from .kl import *
from ..function import qunsigned, qsigned


def init_weight(net_module, qnet_module, method='max'):
    if method == 'max':
        threshold_weight_max(net_module, qnet_module)
    elif method == '3sd':
        threshold_weight_3sd(net_module, qnet_module)
    else:
        raise NotImplementedError()
    qnet_module.weight.data = qsigned(
        qnet_module.weight,
        qnet_module.weight_log2_t.to(qnet_module.weight.device),
        qnet_module.weight_bit_width)


def init_bias(net_module, qnet_module, method='max'):
    if method == 'max':
        threshold_bias_max(net_module, qnet_module)
    elif method == '3sd':
        threshold_bias_3sd(net_module, qnet_module)
    else:
        raise NotImplementedError()
    qnet_module.bias.data = qsigned(
        qnet_module.bias, qnet_module.bias_log2_t.to(qnet_module.bias.device),
        qnet_module.bias_bit_width)


def init_acti(net_module, qnet_module, bin_number=2048, cali_number=128):
    entropy_calibration(net_module,
                        qnet_module,
                        bin_number=bin_number,
                        cali_number=cali_number)


def init_network(net_proc,
                 qnet_proc,
                 name,
                 weight_method='max',
                 bias_method='max',
                 bin_number=2048,
                 cali_number=128,
                 show=False):
    modules = list(net_proc._modules.keys())
    if modules == []:
        if hasattr(qnet_proc, 'acti_log2_t'):
            if qnet_proc.quant is True:
                init_acti(net_proc,
                          qnet_proc,
                          bin_number=2048,
                          cali_number=128)
                if show:
                    print(name, ': activation threshold is quanted as',
                          float(qnet_proc.acti_log2_t))
        if hasattr(qnet_proc, 'weight_log2_t'):
            if qnet_proc.quant is True:
                init_weight(net_proc, qnet_proc, method=weight_method)
                if show:
                    print(name, ': weight threshold is quanted as',
                          float(qnet_proc.weight_log2_t))
        if hasattr(qnet_proc, 'bias_log2_t') and qnet_proc.bias is not None:
            if qnet_proc.quant is True:
                init_bias(net_proc, qnet_proc, method=bias_method)
                if show:
                    print(name, ': bias threshold is quanted as',
                          float(qnet_proc.bias_log2_t))
    else:
        for key in modules:
            init_network(net_proc._modules[key],
                         qnet_proc._modules[key],
                         name + '.' + key,
                         weight_method=weight_method,
                         bias_method=bias_method,
                         bin_number=bin_number,
                         cali_number=cali_number,
                         show=show)



__title__ = 'tqt'
__description__ = 'torch implement of TQT'
__version__ = '0.4.2'
__author__ = 'Pannenets F'
__author_email__ = 'pannenets.f@foxmail.com'
__license__ = 'BSD-3-Clause License'
__url__ = 'https://github.com/PannenetsF/TQT'



from . import function as tqt
import torch.nn as nn

OFUNCTION = {
    'ReLU':
    lambda inplace=False, acti_bit_width=8, retrain=True: nn.ReLU(inplace=
                                                                  inplace),
    'ReLU6':
    lambda inplace=False, acti_bit_width=8, retrain=True: nn.ReLU6(inplace=
                                                                   inplace),
    'BatchNorm2d':
    lambda num_features, eps=1e-5, momentum=0.1, affine=True,
    track_running_stats=True, weight_bit_width
    =8, bias_bit_width=16, retrain=True: nn.BatchNorm2d(
        num_features, eps, momentum, affine, track_running_stats),
    'Conv2d':
    lambda in_channels, out_channels, kernel_size, stride=1, padding=0,
    dilation=1, groups=1, bias=True, padding_mode='zeros', weight_bit_width=8,
    bias_bit_width=16, retrain=True: nn.Conv2d(in_channels,
                                               out_channels,
                                               kernel_size,
                                               stride=stride,
                                               padding=padding,
                                               dilation=dilation,
                                               groups=groups,
                                               bias=bias,
                                               padding_mode=padding_mode),
    'Linear':
    lambda in_features, out_features, bias=True, weight_bit_width=8,
    bias_bit_width=16, retrain=True: nn.Linear(
        in_features, out_features, bias=bias),
    'MaxPool2d':
    lambda kernel_size, stride=None, padding=0, dilation=1, return_indices=
    False, ceil_mode=False: nn.MaxPool2d(kernel_size,
                                         stride=stride,
                                         padding=padding,
                                         dilation=dilation,
                                         return_indices=return_indices,
                                         ceil_mode=ceil_mode),
    'Adder2d':
    lambda input_channel, output_channel, kernel_size, stride=1, padding=0,
    bias=False, weight_bit_width=8, bias_bit_width=16, retrain=True: tqt.extra.
    Adder2d(input_channel,
            output_channel,
            kernel_size,
            stride=stride,
            padding=padding,
            bias=bias)
}

QFUNCTION = {
    'ReLU':
    lambda inplace=False, acti_bit_width=8, retrain=True, quant=False: tqt.
    ReLU(inplace=inplace,
         acti_bit_width=acti_bit_width,
         retrain=retrain,
         quant=quant),
    'ReLU6':
    lambda inplace=False, acti_bit_width=8, retrain=True, quant=False: tqt.
    ReLU6(inplace=inplace,
          acti_bit_width=acti_bit_width,
          retrain=retrain,
          quant=quant),
    'BatchNorm2d':
    lambda num_features, eps=1e-5, momentum=0.1, affine=True,
    track_running_stats=True, weight_bit_width=8, bias_bit_width=16, retrain=
    True, quant=False: tqt.BatchNorm2d(num_features,
                                       eps,
                                       momentum,
                                       affine,
                                       track_running_stats,
                                       weight_bit_width=weight_bit_width,
                                       bias_bit_width=bias_bit_width,
                                       retrain=retrain,
                                       quant=quant),
    'Conv2d':
    lambda in_channels, out_channels, kernel_size, stride=1, padding=0,
    dilation=1, groups=1, bias=True, padding_mode='zeros', weight_bit_width=8,
    bias_bit_width=16, retrain=True, quant=False: tqt.Conv2d(
        in_channels,
        out_channels,
        kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=groups,
        bias=bias,
        padding_mode=padding_mode,
        weight_bit_width=weight_bit_width,
        bias_bit_width=bias_bit_width,
        retrain=retrain,
        quant=quant),
    'Linear':
    lambda in_features, out_features, bias=True, weight_bit_width=8,
    bias_bit_width=16, retrain=True, quant=False: tqt.Linear(
        in_features,
        out_features,
        bias=bias,
        weight_bit_width=weight_bit_width,
        bias_bit_width=bias_bit_width,
        retrain=retrain,
        quant=quant),
    'MaxPool2d':
    lambda kernel_size, stride=None, padding=0, dilation=1, return_indices=
    False, ceil_mode=False, quant=False: nn.MaxPool2d(kernel_size,
                                                      stride=stride,
                                                      padding=padding,
                                                      dilation=dilation,
                                                      return_indices=
                                                      return_indices,
                                                      ceil_mode=ceil_mode),
    'Adder2d':
    lambda input_channel, output_channel, kernel_size, stride=1, padding=0,
    bias=False, weight_bit_width=8, bias_bit_width=16, retrain=True, quant=
    False: tqt.Adder2d(input_channel,
                       output_channel,
                       kernel_size,
                       stride=stride,
                       padding=padding,
                       bias=bias,
                       weight_bit_width=weight_bit_width,
                       bias_bit_width=bias_bit_width,
                       retrain=retrain,
                       quant=False)
}
